require "util"
require "defines"

on_init_calls = {}
on_load_calls = {}

function setup()
	global.data = global.data or ontick:new()
	global.events_manager = global.events_manager or events_manager
	ontick:init()
	events_manager:init()

	ontick:start()
end

script.on_init(setup)
script.on_load(setup)
script.on_configuration_changed(setup)


ontick = {data = "data"}
function ontick:new(o)
	o = o or {}   -- create object if user does not provide one
	setmetatable(o, self)
	self.__index = self
	return o
end

function ontick:init()
	self.__index = self
	setmetatable(global.data, self)
end

function ontick:on_tick(_event)
	game.player.print(self.data .. _event.tick*2)
end

function ontick:start()
	global.events_manager:add_listener(defines.events.on_tick, self)
end



events_manager = {
	names = {},
	registered_listeners = {},
	registered_functions = {}
}

function events_manager:init()
	for event_name,event_id in pairs(defines.events) do
		script.on_event(event_id, function (event) events_manager:event(event) end)
		self.names[event_id] = event_name
	end
end

function events_manager:event(_event)
	local event_name = self.names[_event.name]
	if self.registered_listeners[_event.name] then -- only if the event has a listener
		for listener,fun in pairs(self.registered_listeners[_event.name]) do
			if listener[event_name] then
				if fun == 1 then listener[event_name](listener, _event) end
			end
		end
	end
	if self.registered_functions[_event.name] then
		for id,fun in pairs(self.registered_functions[_event.name]) do
			if type(fun) == "function" then fun(_event, id) end
		end
	end
end

function events_manager:add_listener(_event_id, _listener)
	if not self.names[_event_id] then self:init() end -- if an event is not recorded, reload them all
	-- if after reloading the event is still not recorded
	if not self.names[_event_id] then error("the event must be registered in the defines.events table. If you created the event yourself, please use 'defines.events.<event_name> = generate_event_name()'") end
	local event_name = self.names[_event_id]
	
	if type(_event_id) ~= "number" then error("invalid event id: '" .. _event_id.. "'") end
	if type(_listener) ~= "table"  then error("invalid listener : the listener must be a table (object)") end
	if not _listener[event_name] then error("invalid listener : the listener must conatin a function named " .. event_name) end

	if not self.registered_listeners[_event_id] then self.registered_listeners[_event_id] = {} end
	self.registered_listeners[_event_id][_listener] = 1
end

function events_manager:remove_listener(_event_id, _listener)
	if type(_event_id) ~= "number" then error("invalid event id: '" .. _event_id.. "'") end
	if type(_listener) ~= "table"  then error("invalid listener : the listener must be a table (object)") end
	self.registered_listeners[_event_id][_listener] = nil
end


function events_manager:add_function(_event_id, _function_id, _function)
	if not self.names[_event_id] then self:init() end -- if an event is not recorded, reload them all
	-- if after reloading the event is still not recorded
	if not self.names[_event_id] then error("the event must be registered in the defines.events table. If you created the event yourself, please use 'defines.events.<event_name> = generate_event_name()'") end
	local event_name = self.names[_event_id]
	
	if type(_event_id) ~= "number" then error("invalid event id: '" .. _event_id.. "'") end
	if type(_function_id) ~= "string" then error("invalid argument, _function_id must be a string") end
	if type(_function) ~= "function" then error("invalid argument, _function must have type function") end



	if not self.registered_functions[_event_id] then self.registered_functions[_event_id] = {} end
	self.registered_functions[_event_id][_function_id] = _function
end

function events_manager:remove_function(_event_id, _function_id)
	if type(_event_id) ~= "number" then error("invalid event id: '" .. _event_id.. "'") end
	if type(_function_id) ~= "string" then error("invalid argument, _function_id must be a string") end
	self.registered_functions[_event_id][_function_id] = nil
end




local function print(str,trace)
	if not game then return end
	if not pClosure.printMessages then return end
	game.players[1].print(str)
	if trace then 
		game.players[1].print(debug.traceback())
	end
end